function [negCLL, gradCLL] = OptimizationCLL_TPAMI(beta, A, ns, IndexSet, IndexSet2, start_indices, samples)
% OPTIMIZATIONCLL_TPAMI Calculates the conditional log-likelihood 
%   [negCLL, gradCLL] = OPTIMIZATIONCLL_TPAMI(beta, A, ns, IndexSet, ...
%      IndexSet2, start_indices, samples) calculates the conditional 
%   log-likelihood (CLL) and its gradient with respect to the supplied
%   log-proabilities.
%
%   The input and output parameters are as follows:
%    * beta: log-probabilities (as a vector)
%    * A: adjacency matrix of the Bayesian network to use
%    * ns: vector of sizes of the nodes in network, i.e. number of discrete 
%       values the nodes can take
%    * IndexSet, IndexSet2, start_indices: Precalculated indices for
%       shorter runtimes of the code
%    * samples: matrix of samples (each column is a sample)
%    
%    * negCLL: negative CLL of the training samples under the
%       log-probabilities beta
%    * gradCLL: gradient of the CLL with respect to the log-probabilities
%       beta
%
% --
% The provided software may be used free of charge for research purposes.
% For other uses and further support please contact Franz Pernkopf,
% pernkopf@tugraz.at.
%
% References:
%  F. Pernkopf, M. Wohlmayr, and S. Tschiatschek, 
%  "Maximum Margin Bayesian Network Classifiers," 
%  IEEE Transactions on Pattern Analysis and Machine Intelligence,
%  vol. 99, no. PrePrints, 2011
%
% Authors:
%  F. Pernkopf, M. Wohlmayr, and S. Tschiatschek
%
% Version:
%  1.0 August 24th, 2011
%

% Turn off Division-By-Zero Warnings
warning off MATLAB:divideByZero

%% Define frequently used quantities
nrSamples = size(samples, 2);
nrNodes = length(ns);
nrClasses = ns(1);

%% renormalize probabilities
theta = exp(beta);
theta = normalize(A, ns, theta);

%% Compute negCLL(beta)

% Compute sample probabilities using the current CPDs
JointProb = ones(nrSamples,nrClasses);
for j = 1:nrNodes,
    for c = 1:ns(1),
        JointProb(:,c) = JointProb(:,c) .* theta(IndexSet(:,j,c));
    end
end

%% Compute W
W = zeros(size(JointProb));
for m = 1:nrSamples, % cycle through the samples
    W(m,:) = JointProb(m,:) ./ sum(JointProb(m,:));
end

if any(~isfinite(W)),
    warning('CLLLearning:InfiniteElements', 'There are non-finite Elements');
    W(isnan(W)) = 0;
    W(~isfinite(W)) = 1e5;
end

%% Calculate objective
CLL = 0;
for m=1:nrSamples,
    CLL = CLL + log(JointProb(m,samples(1,m)) / sum(JointProb(m,:)));
end
negCLL = -CLL; % objective (to be minimized)

%% Compute the gradient gradCLL(beta)
gradCLL = zeros(length(theta),1);
sidx = 1;
for j = 1:nrNodes,
    fam = [j, family(A,j)];
    sizes = ns(fam);
    nrParentStates = prod(sizes(2:end));
    
    indices = subv2ind(sizes', samples(fam,:)'); % each row of data' is a case

    count = zeros(prod(sizes),1);
    count(1:max(indices)) = accumarray(indices, ones(1,nrSamples));
    countsA = myreshape(count, sizes);
        
    if j == 1, % class node
        countsB = sum(W)';
    else % non-class nodes
        countsB=zeros(size(count));
        for z = 1:nrClasses,
            indi = IndexSet2(:, j, z);  
            countsB(1:max(indi)) = countsB(1:max(indi)) + accumarray(indi, W(:,z));
        end
    end
    countsB = myreshape(countsB, sizes);

    Term1 = reshape(countsA - countsB, sizes(1), nrParentStates);
    
    Params = reshape(theta(start_indices(j):(start_indices(j+1)-1)), sizes(1), nrParentStates);

    Term2 = (ones(size(Term1,1),1)*sum(Term1,1)).*(-1*Params);
    BG = Term1 + Term2;
    
    gradCLL(sidx:sidx + numel(BG) - 1) = reshape(-1*BG,numel(BG),1);

    sidx = sidx + numel(BG);
end

end

