function [theta] = learnCGMMParameters(A, ns, theta_initial, samples, iterations, kappa, lambda)
%LEARNCGCMMPARAMETERS Learns network parameters by maximizing the margin
%  [theta] = LEARNCGCLLPARAMETERS(A, ns, theta_initial, samples, ...
%     iterations) learns network parameters by maximizing the margin
%     using a gradient based method.
%
%   The input and output parameters are as follows:
%    * A: adjacency matrix of the Bayesian network to use
%    * ns: vector of sizes of the nodes in network, i.e. number of discrete 
%       values the nodes can take
%    * theta_initial: initial probabilities (starting point for
%      optimization)
%    * samples: matrix of samples (each column is a sample)
%    * iterations: maximum number of conjugate gradient steps the algorithm
%       is allowed to perform
%    * kappa/lambda: tuning parameters (see reference for details)
%
%    * theta: probabilities that maximize the margin of the samples in the
%       given Bayesian network
%
% --
% The provided software may be used free of charge for research purposes.
% For other uses and further support please contact Franz Pernkopf,
% pernkopf@tugraz.at.
%
% References:
%  F. Pernkopf, M. Wohlmayr, and S. Tschiatschek, 
%  "Maximum Margin Bayesian Network Classifiers," 
%  IEEE Transactions on Pattern Analysis and Machine Intelligence,
%  vol. 99, no. PrePrints, 2011
%
% Authors:
%  F. Pernkopf, M. Wohlmayr, and S. Tschiatschek
%
% Version:
%  1.0 August 24th, 2011
%

%% reshape parameters
ns = ns(:);
theta_initial = theta_initial(:);

%% parameters
epsilon = 1e-5;

%% Parameters of the given Bayesian network
nrNodes = length(ns);
nrClasses = ns(1);
nrSamples = size(samples, 2);

%% (1) Initialize betas
if isempty(theta_initial), % are initial values supplied?
    theta = learnMLParameters(A, ns, samples); % No -> determine ML parameters
else
    theta = theta_initial;
end

% are there any zero probabilites?
if any(theta == 0),
    theta(theta == 0) = epsilon;
    
    theta = normalize(A, ns, theta);
end

beta = log(theta);

%% (2) Precalculate often used quantities
% Determine for every sample the set of classes to which the sample does
% not belong
classes = 1:nrClasses;
sset = zeros(nrClasses-1, nrSamples);
for sidx = 1:nrSamples,
    sset(:, sidx) = setdiff(classes, samples(1, sidx));
end

% Create array with start and end indices of CPTs for certain variables
% over all parent states
start_indices = zeros(nrNodes+1,1);
sidx = 1;
eidx = 0;
for k=1:nrNodes,
    if k > 1,
        sidx = eidx + 1;
    end

    parents = logical(A(:,k));
    nrParentStates = prod(ns(parents));

    eidx = eidx + ns(k) * nrParentStates;
    
    start_indices(k) = sidx;
end
start_indices(end) = eidx + 1;

% Precalculate indices in joint probability matrix
IndexSet = zeros(nrSamples,nrNodes, nrClasses);
IndexSet2 = zeros(nrSamples,nrNodes, nrClasses);
sidx = 0;
for j = 1:nrNodes,
    for c = 1:ns(1),
        fam = [j, family(A,j)];
        
        if length(fam) == 1,
            data = c*ones(nrSamples, 1);
        else
            data = samples(fam, :)';
            
            % replace class
            if ~isempty(find(fam == 1, 1)),
                data(:, fam == 1) = c*ones(nrSamples,1);
            end
        end
        
        % determine vector b
        b = subv2ind(ns(fam)', data);
        
        % Prob(:,j,c)=x(b);
        IndexSet(:,j,c) = sidx + b;
        IndexSet2(:,j,c) = b;
    end
    
    sidx = sidx + prod(ns(fam(1:end)));
end

%% (3) Perform Maximum-Margin learning
beta = minimize(beta, 'OptimizationMM_TPAMI', -iterations, 1, A, ns, IndexSet, IndexSet2, start_indices, samples, sset, kappa, lambda);

%% (4) renormalize obtained probabilities
theta = exp(beta);
theta = normalize(A, ns, theta);
    
end
